//
// Msql client library for JAVA
// Copyright (C) 1995 Darryl Collins <darryl@minmet.uq.oz.au>
// This software is provided "as is" without any expressed or implied warranty.
//
// This package implements a client for the mSQL server designed and
// implemented by David Hughes <bambi@hughes.com.au>
//
// Revision 1.2  1997/02/25
//

package msql;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;


/**
 * Msql Class - implements a connection to a remote server
 * and a JAVA variant of the Msql API as defined by David Hughes
 *
 * @version 1.2.0
 * @author <a href="http://www.minmet.uq.oz.au/~darryl">Darryl Collins</a>
 */
public class Msql {

	// Communication variables
	private Socket s;
	private DataInputStream in;
	private DataOutputStream out;

	// User name and version info
	private String userName;
	private String strVersion;
	private int iMsqlVersion;
	private int iProtocolVersion;

	// Temporary storage for reply packets
	private String strReply;

	// TCP port
	private int msqlPort = 1114;

	// Type constants
	public static final int INT=1;
	public static final int CHAR=2;
	public static final int REAL=4;
	public static final int IDENT=8;
	public static final int NULL=16;
	public static final int TEXT=32;
	public static final int UINT=64;
	public static final int IDX=253;
	public static final int SYSVAR=254;
	public static final int ANY_TYPE=254;


/**
 * Constructor
 */
	public Msql() {
	}

/**
 * Msql Constructor - establishes connection to named host
 *
 * @param msqlServer	the name of the msql server.
 */
	public Msql(String msqlServer) throws MsqlException {

		if(msqlServer != null) 	
			Connect(msqlServer);
	}

/**
 * Connect - establishes connection to named host
 *
 * @param msqlServer	the name of the msql server.
 * @param userName	user name the applet will assume.
 * @exception 		MsqlException
 *
 * 	1. Open socket.
 *	2. Establish IO Streams on socket.
 *	3. Obtain version info.
 *	4. Supply user name.
 *	5. Retrieve status.
 */
	public synchronized void Connect(String msqlServer, String userName) throws MsqlException {

		try {
			// Open connection 
			s = new Socket(msqlServer, msqlPort);

			// Get Streams
			in = new DataInputStream(s.getInputStream());
			out = new DataOutputStream(s.getOutputStream());

			// Read Version Info 
			strVersion = recvString();
			if(strVersion.startsWith("-")) {
				throw new MsqlException(
			  		strVersion
				);
			}

			// Parse Version String
			int i = strVersion.indexOf(':');
			if (i == -1) {
				throw new MsqlException(
			  		strVersion
				);
			}
			int j = strVersion.indexOf(':',i+1);
			if (j == -1) {
				throw new MsqlException(
			  		strVersion
				);
			}

			try {
				iMsqlVersion = 
				 Integer.parseInt(strVersion.substring(0,i));
				iProtocolVersion = 
				 Integer.parseInt(strVersion.substring(i+1,j));
			} catch(NumberFormatException e) {
				throw new MsqlException(
			  		"Number format exception in version string: "+strVersion
				);
			}

			// Send Username
			this.userName = userName;
			sendString(userName);

			// Get reply
			strReply = recvString();
			if(strReply.startsWith("-1:",0))
				throw new MsqlException("Msql connection failed: "+strReply);

		} catch(Exception e) {
			throw new MsqlException(
				"Msql.Connect: "+e
			);
		}
	}

/**
 * Connect - connect using a specified port
 *
 * @param msqlServer	the name of the msql server.
 * @param userName	user name the applet will assume.
 * @param tcpPort	port for msql connections
 * @exception		MsqlException as above.
 */
	public void Connect(String msqlServer, String userName, int tcpPort) 
							throws MsqlException {

		msqlPort = tcpPort;

		Connect(msqlServer, userName);
	}

/**
 * Connect - establishes connection to named host
 *
 * @param msqlServer	the name of the msql server.
 * @exception		MsqlException as above.
 */
	public void Connect(String msqlServer) throws MsqlException {

		// try {
		// 	String userName = System.getProperty("user.name"));
		// 	Connect(msqlServer, userName);
		// } catch(SecurityException e) {
			Connect(msqlServer, "nobody");
		// }
	}

/**
 * Connect - connect using a specified port
 *
 * @param msqlServer	the name of the msql server.
 * @param tcpPort	port for msql connections
 * @exception		MsqlException as above.
 */
	public void Connect(String msqlServer, int tcpPort) throws MsqlException {

		// try {
		// 	String userName = System.getProperty("user.name"));
		// 	Connect(msqlServer, userName, tcpPort);
		// } catch(SecurityException e) {
			Connect(msqlServer, "nobody", tcpPort);
		// }
	}

/**
 * Sends a String as an mSQL packet.
 *
 * @param str		the string to send to the server.
 * @exception		IOException
 */
	private void sendString(String str) throws IOException {
		
		int i,j,k;

		j = k = str.length();

		out.writeByte(k); k >>>= 8;
		out.writeByte(k); k >>>= 8;
		out.writeByte(k); k >>>= 8;
		out.writeByte(k);
		
		for(i=0; i<j; i++) {
			byte b = (byte)str.charAt(i);
			out.writeByte(b);
		}
	}

/**
 * Receives a packet and converts it to a String
 *
 * @return 		the string retrieved from the socket.
 * @exception		IOException
 */
	private String recvString() throws IOException {

		StringBuffer str = new StringBuffer();

		int i,j,k=0;

		j = in.readUnsignedByte(); k += j;
		j = in.readUnsignedByte(); k += j<<8;
		j = in.readUnsignedByte(); k += j<<16;
		j = in.readUnsignedByte(); k += j<<24;

		for(i=0; i<k; i++) {
			char c = (char)in.readByte();
			str.append(c);
		}

		return str.toString();
	}

/**
 * Close - close the connection to the server.
 *
 * @exception		MsqlException
 */
	public synchronized void Close() throws MsqlException {

		try {
			sendString("1:\n");
			s.close();
		} catch(Exception e) {
			throw new MsqlException("Msql.Quit: "+e);
		}
	}

/**
 * SelectDB - Select the database to work with
 * 
 * @param db		the name of the database to select.
 * @exception		MsqlException if select fails. 
 */
	public synchronized void SelectDB(String db) throws MsqlException {

		try {
			sendString("2:"+db+"\n");
			strReply = recvString();
			if(strReply.startsWith("-1:",0))
				throw new MsqlException("Msql.SelectDB: "+strReply.substring(3));

		} catch(Exception e) {
			throw new MsqlException("Msql.SelectDB: "+e);
		}
	}

/**
 * Send an SQL Query to the server.  
 *
 * @param s		the Query string.
 * @return		null when used with INSERT/UPDATE or similar queries which do not return data
 * @return		instance of MsqlResult when used with SELECT.
 * @exception		MsqlException when query or communication fails.
 */
	public synchronized MsqlResult Query(String s) throws MsqlException {

		int m,n;
		MsqlData table, fields;
		MsqlResult results;

		try {
			// Send the command
			sendString("3:"+s+"\n");
			
			// Retrieve the reply
			strReply = recvString();

		} catch(IOException e) {
			throw new MsqlException("Msql.Query: "+e);
		}

		// Check for errors
		if(strReply.startsWith("-1:"))
			throw new MsqlException("Msql.Query: "+strReply.substring(2));
		// Determine number of fields
		if(strReply.length() < 2) 
			return null;

		m = strReply.indexOf(':',2);
		if(m == -1)
			return null;

		n=Integer.parseInt(strReply.substring(2,m));

		if (n <= 0)
			return null;

		// Receive the row data
		table = recvData(n);

		// Receive the field data
		fields = recvData(6);

		results = new MsqlResult(table, fields);

		return results;
	}

/**
 * List Databases available on server. 
 *
 * @return		MsqlData structure cointaining names of databases.
 * @exception		MsqlException if ListDB or communication fails.
 */
	private synchronized MsqlData _ListDBs() throws MsqlException {

		MsqlData result;

		try {
			sendString("4:\n");
		} catch (Exception e) {
			throw new MsqlException("Msql._ListDBs: "+e);
		}

		result = recvData(1);

		return result;
	}

/**
 * List Databases available on server. 
 *
 * @return		array of String cointaining names of databases.
 * @exception		MsqlException if ListDB or communication fails.
 */
	public synchronized String[] ListDBs() throws MsqlException {

		MsqlData result;
		String dbList[];

		result = _ListDBs();

		int n = result.NumRows();

		if(n != 0) {
			dbList = new String[n];
			for(int i=0; i<n; i++)
				dbList[i] = new String((result.FetchRow())[0]);
			return dbList;
		}

		return null;
	}

/**
 * List tables in the selected database.
 *
 * @return		MsqlData structure containing names of tables.
 * @exception		MsqlException if ListTables or communication fails.
 */
	private synchronized MsqlData _ListTables() throws MsqlException {

		MsqlData result;

		try {
			sendString("5:\n");
		} catch (Exception e) {
			throw new MsqlException("Msql._ListTables: "+e);
		}

		result = recvData(1);

		return result;
	}
/**
 * List tables in the selected database.
 *
 * @return		array of String conttaining names of tables.
 * @exception		MsqlException if ListTables or communication fails.
 */
	public synchronized String[] ListTables() throws MsqlException {

		MsqlData result;
		String tableList[];

		result = _ListTables();
 
		int n = result.NumRows();

		if(n != 0) {
			tableList = new String[n];
			for(int i=0; i<n; i++) 
				tableList[i] = new String((result.FetchRow())[0]);
			return tableList;
		}
				
		return null;
	}

/**
 * List Fields in a given table.
 *
 * @param s		name of the table for which information requested.
 * @return		MsqlData structure containing field information.
 * @exception		MsqlException if ListTables or communication fails.
 */
	public synchronized MsqlData _ListFields(String s) throws MsqlException {

		MsqlData result;

		try {
			sendString("6:"+s+"\n");
		} catch (Exception e) {
			throw new MsqlException("Msql._ListFields: "+e);
		}

		result = recvData(6);

		return result;
	}

/**
 * List Fields in a given table.
 *
 * @param s		name of the table for which information requested.
 * @return		array of MsqlFieldDesc containing field information.
 * @exception		MsqlException if ListTables or communication fails.
 */
	public synchronized MsqlFieldDesc[] ListFields(String s) throws MsqlException {

		MsqlData result;
		MsqlFieldDesc fieldList[];

		result = _ListFields(s);

		int n = result.NumRows();

		if(n != 0) {
			fieldList = new MsqlFieldDesc[n];
			for(int i=0; i<n; i++) 
				fieldList[i] = new MsqlFieldDesc(result.FetchRow());
			return fieldList;
		}

		return null;
	}
			

/**
 * Create a new database.  Included for completeness only since
 * command can only be executed if issued from UNIX domain socket.
 */
	public synchronized void CreateDB(String s) throws MsqlException {

		try {
			sendString("7:"+s+"\n");
			strReply = recvString();
			if(strReply.startsWith("-1:",0))
				throw new MsqlException("Msql.CreateDB: "+strReply.substring(3));

		} catch(Exception e) {
			throw new MsqlException("Msql.CreateDB: "+e);
		}
	}

/**
 * Drop a  database.  Included for completeness only since
 * command can only be executed if issued from UNIX domain socket.
 */
	public synchronized void DropDB(String s) throws MsqlException {

		try {
			sendString("8:"+s+"\n");
			strReply = recvString();
			if(strReply.startsWith("-1:",0))
				throw new MsqlException("Msql.DropDB: "+strReply.substring(3));

		} catch(Exception e) {
			throw new MsqlException("Msql.DropDB: "+e);
		}
	}

/**
 * Reload access control list.  Included for completeness only since
 * command can only be executed if issued from UNIX domain socket.
 */
	public synchronized void ReloadACL(String s) throws MsqlException {

		try {
			sendString("9:\n");
			strReply = recvString();
			if(strReply.startsWith("-1:",0))
				throw new MsqlException("Msql.ReloadACL: "+strReply.substring(3));

		} catch(Exception e) {
			throw new MsqlException("Msql.ReloadACL: "+e);
		}
	}

/**
 * Shutdown server.  Included for completeness only since
 * command can only be executed if issued from UNIX domain socket.
 */
	public synchronized void Shutdown() throws MsqlException {

		try {
			sendString("10:\n");
			strReply = recvString();
			if(strReply.startsWith("-1:",0))
				throw new MsqlException("Msql.Shutdown: "+strReply.substring(3));

		} catch(Exception e) {
			throw new MsqlException("Msql.Shutdown: "+e);
		}
	}

/**
 * Retrieve the results of a query.  Returns an instance of MsqlData.
 * Retrieves pakcets from server until an error status (-1:) or
 * success status (-100:) is received.
 */
	private MsqlData recvData(int n) throws MsqlException {

		MsqlData result = new MsqlData(n);

		boolean finished = false;

		while(!finished) {
			try {
				strReply = recvString();
			} catch(IOException e) {
				throw new MsqlException("Msql.recvData: "+e);
			}

			if(strReply.startsWith("-1:",0))
				throw new MsqlException("Msql.recvData: "+strReply.substring(3));

			if(strReply.startsWith("-100:",0))
				finished = true;
			else
				result.addRow(strReply);
		}

		return result;
	}
}
