package security.srp;

import java.io.DataInputStream;
import java.io.IOException;
import security.math.BigInteger;
import security.crypt.SHA;
import security.crypt.Util;

public class SRPClient {
  private BigInteger N;
  private BigInteger g;
  private byte[] s;
  private BigInteger xs;
  private BigInteger xp;
  private BigInteger ws;
  private BigInteger wp;
  private String user;
  private byte[] key;
  private SHA hash, ckhash;

  private static int WS_LEN = 8;

  public SRPClient(String username, byte[] modulus, byte[] generator,
		   byte[] salt) {
    user = username;
    N = new BigInteger(modulus);
    g = new BigInteger(generator);
    s = salt;
    key = null;
    hash = new SHA();
    hash.add(Util.xor(SHA.hash(modulus), SHA.hash(generator), 20));
    hash.add(SHA.hash(username));
    hash.add(salt);
    ckhash = new SHA();
  }

  public byte[] modulus() { return N.toByteArray(); }
  public byte[] generator() { return g.toByteArray(); }
  public byte[] salt() { return s; }
  public byte[] exponential() { return wp.toByteArray(); }
  public byte[] sessionKey() {  return key; }

  public byte[] generateExponential() {
    do {
      ws = new BigInteger(Util.RNG.nextBytes(WS_LEN));
    } while(ws.cmp(BigInteger.one) <= 0);
    wp = new BigInteger(g).modExp(ws, N);
    byte[] out = wp.toByteArray();
    hash.add(out);
    ckhash.add(out);
    return out;
  }

  public void inputPassword(String pass) {
    //xs = new BigInteger().mul(new BigInteger(Util.userHash(user, pass)), s);
    SHA ctxt = new SHA();
    ctxt.add(s);
    ctxt.add(Util.userHash(user, pass));
    xs = new BigInteger(ctxt.digest());
    xp = new BigInteger(g).modExp(xs, N);
  }

  public byte[] getSessionKey(byte[] srvexp) {
    hash.add(srvexp);
    byte[] uhash = SHA.hash(srvexp);
    byte[] fourbytes = {uhash[0], uhash[1], uhash[2], uhash[3]};
    BigInteger sum =
      new BigInteger().mul(new BigInteger(fourbytes), xs);
    BigInteger base = new BigInteger(srvexp);
    if(base.cmp(xp) < 0)
      base.add(base, N);
    base.sub(base, xp);
    BigInteger S = base.modExp(new BigInteger().add(sum, ws), N);
    key = Util.sessionKeyHash(S.toByteArray());
    hash.add(key);
    return key;
  }

  public byte[] response() {
    byte[] resp = hash.digest();
    ckhash.add(resp);
    ckhash.add(key);
    return resp;
  }

  public boolean verify(byte[] resp) {
    return Util.matches(resp, ckhash.digest());
  }

  public static void main(String[] args) {
    SRPClient cli = null;
    DataInputStream stdin = new DataInputStream(System.in);
    String u = null;

    try {
      System.out.print("Enter username: ");
      System.out.flush();
      u = stdin.readLine();
      String str;
      System.out.print("Enter N (from server): ");
      System.out.flush();
      str = stdin.readLine();
      byte[] N = Util.fromb64(str);
      System.out.print("Enter g (from server): ");
      System.out.flush();
      str = stdin.readLine();
      byte[] g = Util.fromb64(str);
      System.out.print("Enter salt (from server): ");
      System.out.flush();
      str = stdin.readLine();
      byte[] salt = Util.fromb64(str);
      
      cli = new SRPClient(u, N, g, salt);

      byte[] ex = cli.generateExponential();
      System.out.println("Wp (to server): " + Util.tob64(ex));

      System.out.print("Enter Yp (from server): ");
      System.out.flush();

      str = stdin.readLine();

      String password =
	new PasswordEntryBox("Enter Password", "Please enter SRP password").
	getAnswer();
      if(password == null)
	System.exit(1);

      cli.inputPassword(password);
      byte[] key = cli.getSessionKey(Util.fromb64(str));
      System.out.println("Session key: " + Util.tohex(key));

      System.out.println("Response (to server): " + Util.tohex(cli.response()));
    }
    catch(IOException e) {
      e.printStackTrace();
    }
    finally {
      System.exit(0);
    }
  }
}
