package security.srp;

import java.io.DataInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import security.math.BigInteger;
import security.crypt.Util;
import security.crypt.SHA;

public class SRPServer {
  private BigInteger N;
  private BigInteger g;
  private BigInteger xp;
  private byte[] s;
  private BigInteger ys;
  private BigInteger yp;
  private byte[] key;
  private SHA hash, ckhash;

  private static int YS_LEN = 8;

  public SRPServer(String username, PasswordFile pw)
       throws NoSuchUserException {
    String[] result = pw.lookup(username);
    if(result == null)
      throw new NoSuchUserException();
    xp = new BigInteger(Util.fromb64(result[0]));
    s = Util.fromb64(result[1]);
    byte[] nb = Util.fromb64(result[3]);
    byte[] gb = Util.fromb64(result[4]);
    g = new BigInteger(gb);
    N = new BigInteger(nb);
    hash = new SHA();

    ckhash = new SHA();
    ckhash.add(Util.xor(SHA.hash(nb), SHA.hash(gb), 20));
    ckhash.add(SHA.hash(username));
    ckhash.add(s);
    key = null;
  }

  public byte[] modulus() { return N.toByteArray(); }
  public byte[] generator() { return g.toByteArray(); }
  public byte[] salt() { return s; }
  public byte[] exponential() { return yp.toByteArray(); }
  public byte[] sessionKey() { return key; }

  public byte[] generateExponential() {
    do {
      ys = new BigInteger(Util.RNG.nextBytes(YS_LEN));
    } while(ys.cmp(BigInteger.one) <= 0);
    yp = new BigInteger().add(xp, g.modExp(ys, N));
    if(yp.cmp(N) >= 0)
      yp.sub(yp, N);
    return yp.toByteArray();
  }

  public byte[] getSessionKey(byte[] cliexp) {
    ckhash.add(cliexp);
    ckhash.add(yp.toByteArray());
    hash.add(cliexp);
    byte[] uhash = SHA.hash(yp.toByteArray());
    byte[] fourbytes = {uhash[0], uhash[1], uhash[2], uhash[3]};
    BigInteger S = new BigInteger();
    BigInteger.modMul(S, new BigInteger(cliexp),
		      new BigInteger(xp).modExp(new BigInteger(fourbytes), N),
		      N);
    key = Util.sessionKeyHash(S.modExp(ys, N).toByteArray());
    ckhash.add(key);
    return key;
  }

  public byte[] response() {
    return hash.digest();
  }

  public boolean verify(byte[] resp) {
    if(Util.matches(resp, ckhash.digest())) {
      hash.add(resp);
      hash.add(key);
      return true;
    }
    else
      return false;
  }

  public static void main(String[] args) {
    SRPServer serv = null;
    DataInputStream stdin = new DataInputStream(System.in);
    String u = null;

    try {
      PasswordFile pf;
      if(args.length < 1)
	pf = new PasswordFile();
      else
	pf = new PasswordFile(args[0]);
      System.out.print("Enter username: ");
      System.out.flush();
      u = stdin.readLine();
      serv = new SRPServer(u, pf);
    }
    catch(FileNotFoundException e) {
      System.err.println("Password file not found");
      System.exit(1);
    }
    catch(NoSuchUserException e) {
      System.err.println("User " + u + " unknown");
      System.exit(1);
    }
    catch(Exception e) {
      e.printStackTrace();
      System.exit(1);
    }

    System.out.println("N (to client): " + Util.tob64(serv.modulus()));
    System.out.println("g (to client): " + Util.tob64(serv.generator()));
    System.out.println("salt (to client): " + Util.tob64(serv.salt()));

    byte[] ex = serv.generateExponential();

    System.out.print("Enter Wp (from client): ");
    System.out.flush();
    try {
      String wpstr = stdin.readLine();

      // Must get Wp first before revealing Yp/Z
      System.out.println("Yp (to client): " + Util.tob64(ex));

      byte[] key = serv.getSessionKey(Util.fromb64(wpstr));
      System.out.println("Session key: " + Util.tohex(key));

      System.out.print("Enter response (from client): ");
      System.out.flush();
      String resp = stdin.readLine();

      if(serv.verify(Util.fromhex(resp)))
	System.out.println("Authentication successful.");
      else
	System.out.println("Authentication failed.");
    }
    catch(IOException e) {
      e.printStackTrace();
    }
  }
}
