package injektilo.net;

import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;

/**
 * This class implements the Post Office Protocol Version 3 as
 * specified by RFC-1939.
 *
 * <p>Copyright (c) 1998 Jason Diamond</p>
 *
 * <p>sendangels@usa.net<br>
 * http://www.geocities.com/ResearchTriangle/Thinktank/8343/</p>
 *
 * @author Jason Diamond
 * @version 1.1.2
 * 
 * History:
 *
 * 4-Nov-1998 v1.1.1
 *  initial release
 * 18-Nov-1998 v1.1.2
 *  added readTo method
 *  removed retrieve(int, OutputStream)
 *  added private octets and octetsVector attributes
 *  added getOctets and getOctetsVector methods
 *  changed both list methods so that they return a boolean indicating success
 *  added private uid and uidVector attributes
 *  added getUid and getUidVector methods
 *  changed both uidl methods so that they return a boolean indicating success
 *
 */

public class Pop3Connection {
	private Socket socket;
	private BufferedReader in;
	private PrintWriter out;
	
  public static final int DEFAULT_PORT = 110;
  
  private int state;

  private static final int NONE = 0;
  private static final int AUTHORIZATION = 1;
  private static final int TRANSACTION = 2;
  private static final int UPDATE = 3;
      
  private boolean connected;

  private String response;
  
  private String nonce;
  
  private int authentication = AUTO;
  
  /**
   * Constant passed to <CODE>setAuthentication</CODE>. When
   * authentication is set to AUTO, the <CODE>authenticate</CODE>
   * method will first attempt to authenticate the user using the
   * APOP command only if a timestamp was presented when the
   * connection was opened. If not, or if the APOP fails for 
   * some reason, the USER/PASS commands are used.
   *
   * @see #setAuthentication
   * @see #authenticate
   */
   
   public static final int AUTO = 0;   
   
  
  /**
   * Constant passed to <CODE>setAuthentication</CODE>. This is the
   * default value.
   *
   * @see #setAuthentication
   * @see #authenticate
   */
   
  public static final int USER = 1;
  
  /**
   * Constant passed to <CODE>setAuthentication</CODE>. Set it to this
   * value if you want to use the APOP command for authenticating the
   * user's identity. Without sending their password in the clear.
   * Note that most servers don't support this option.
   *
   * @see #setAuthentication
   * @see #authenticate
   */
   
  public static final int APOP = 2;
   
  private int totalMessages = -1;
  private int totalOctets = -1;
  
  private int octets;
  private Vector octetsVector;
  
  private String uid;
  private Vector uidVector;

	private boolean debug;
  private PrintWriter debugOut;

	public Pop3Connection() {
  }
	
	private void sendCommand(String s)
      throws IOException {
    out.print(s + "\r\n");
    out.flush();
    
    if (debug) {
      debugOut.println("C: " + s);
    }
  }
  
  private String getLine()
      throws IOException {
    response = in.readLine();
    
    if (debug) {
      debugOut.println("S: " + response);
    }
    
    return response;
  }
  
  private boolean getResponse()
      throws IOException {
    response = in.readLine();
    
    if (debug) {
      debugOut.println("S: " + response);
    }

    return (response.startsWith("+OK"));
  }
  
  /**
   * Sets the authentication mechanism to use for this connection.
   * Defaults to USER.
   *
   * @param auth the authentication mechanism to use
   * @see #USER
   * @see #APOP
   * @see #getAuthentication
   */
  
  public void setAuthentication(int auth) {
    authentication = auth;
  }
  
  /**
   * Gets the authentication mechanism being used for this connection.
   *
   * @return the currently selected authentication mechanism
   * @see #USER
   * @see #APOP
   * @see #setAuthentication
   */
   
   public int getAuthentication() {
     return authentication;
   }
	
	/**
   * Open a connection to the specified host and port.
   * 
   * @param host the domain name of the POP3 server
   * @param port the port to connect to on the host
   * @return true if connection successful, false if failed
   * @exception UnknownHostException if the host could not be found
   * @exception IOException if an input or output exception occurred
   * @see #open(String)
   */
  
  public boolean open(String host, int port) 
      throws UnknownHostException, IOException {		
    socket = new Socket(host, port);
		in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		out = new PrintWriter(new OutputStreamWriter(socket.getOutputStream()));
    
    connected = true;
    
    nonce = null;
    if (getResponse()) {
      int i = response.indexOf("<");
      if (i != -1) {
        nonce = response.substring(i, response.indexOf(">") + 1);
        if (debug) {
          System.out.println("nonce: " + nonce);
        }
      }
      return true;
    }
    		
		return false;
	}
  
  /**
   * Open a connection to the specified host using the default port.
   *
   * @param host the domain name of the POP3 server
   * @return true if connection successful, false if failed
   * @exception UnknownHostException if the host could not be found
   * @exception IOException if an input or output exception occurred
   * @see #open(String, int)
   */
  
  public boolean open(String host)
      throws UnknownHostException, IOException {      
    return open(host, DEFAULT_PORT);
  }
  
  /**
   * Obtain authorization to access a mailbox. The connection must have
   * already been opened using the <CODE>open</CODE> method. The
   * authentication mechanism (USER/PASS or APOP) is set using the 
   * <CODE>setAuthentication</CODE> method. 
   *
   * @param user the name of the account
   * @param pass the account's password
   * @return true if authorization was granted, false if not
   * @exception IOException if an input or output exception occurred
   * @see #open
   * @see #setAuthentication
   */
	
	public boolean authenticate(String user, String pass)
      throws IOException	{
    switch (authentication) {
      case AUTO:
        if (apop(user, pass)) {
          return true;
        }
      case USER:
        return userPass(user, pass);
      case APOP:
        return apop(user, pass);        
    }
    return false;
  }
  
  private boolean userPass(String user, String pass)
      throws IOException {
    sendCommand("USER " + user);
    if (getResponse()) {
      sendCommand("PASS " + pass);
      if (getResponse()) {
        state = TRANSACTION;
        return true;
      }
    }
    return false;
  }
  
  private boolean apop(String user, String pass)
      throws IOException {
    if (nonce != null) {
      try {
        MessageDigest md = MessageDigest.getInstance("MD5");
        md.update(nonce.getBytes());
        if (pass == null) {
          pass = "";
        }
        byte[] digest = md.digest(pass.getBytes());
        sendCommand("APOP " + user + " " + digestToString(digest));
        if (getResponse()) {
          return true;
        }
      } catch (NoSuchAlgorithmException e) {
        debugOut.println("NoSuchAlgorithmException: " + e.getMessage());
      }
    }
    return false;
  }
  
  /**
   * Check the status of a user's maildrop. This sends the STAT
   * command to the POP3 server. If the command succeeds, you can
   * retrieve the number of messages available and the total size
   * in octets with <CODE>getTotalMessages</CODE> and
   * <CODE>getTotalOctets</CODE>.
   *
   * @return true if successful, false if not
   * @exception IOException if an input or output exception occurred
   * @see #getTotalMessages
   * @see #getTotalOctets
   */
	
	public boolean status()
      throws IOException {
    totalMessages = -1;
    totalOctets = -1;
    sendCommand("STAT");
    if (getResponse()) {		
      int status[] = new int[2];
      status = parseStatus(response.substring(4));
      totalMessages = status[0];
      totalOctets = status[1];
      return true;
    } else {
      return false;
    }
	}
  
  private int[] parseStatus(String s) {
    int status[] = new int[2];
    try {
      status[0] = Integer.parseInt(s.substring(0, s.indexOf(' ')));
      s = s.substring(s.indexOf(' ') + 1);
      int i = s.indexOf(' ');
      if (i != -1) {
        s = s.substring(0, i);
      }
      status[1] = Integer.parseInt(s);
    } catch (NumberFormatException e) {
      debugOut.println("NumberFormatException: " + e.getMessage());
    }
    return status;
  }   
	
	/**
   * Get the total number of messages available for retrieval.
   * The <CODE>status</CODE> method needs to be invoked before
   * this method will return any useful value.
   *
   * @return the number of messages available for retrieval or
   * -1 if the <CODE>status</CODE> method hasn't been called or
   * an error occured when it was called
   * @see #status
   */
   
   public int getTotalMessages() {
		return totalMessages;
	}
	
	/**
   * Get the total size in octets of all the available messages.
   * The <CODE>status</CODE> method needs to be invoked before
   * this method will return any useful value.
   *
   * @return the size in octets of all the available messages
   * or -1 is the <CODE>status</CODE> method hasn't been called
   * or an error occured when it was
   * @see #status
   */
  
  public int getTotalOctets() {
		return totalOctets;
	}
  
  /**
   * Return a Vector containing the size of each available
   * message in octets. The Vector will contain as many elements
   * as the highest numbered message on the server. Messages marked
   * for deletion will be represented by an Integer object with a
   * value of -1. For example, if a session started out with 3
   * messages availble and message 2 was deleted, the resulting
   * vector would contain [message 1's size, -1, message 3's size].
   *
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #getOctetsVector
   */

	public boolean list()
      throws IOException {
    octetsVector = new Vector();
    sendCommand("LIST");
    if (getResponse()) {
      String s;
      int status[] = new int[2];
      int i = 1;
      do {
        s = getLine();
        if (!s.equals(".")) {
          status = parseStatus(s);
          if (status[0] != i) {
            for (int j = i; j < status[0]; ++j) {
              octetsVector.addElement(new Integer(-1));
            }          
          }
          octetsVector.addElement(new Integer(status[1]));
          i = status[0] + 1;
        }
      } while (!response.equals("."));
      return true;
    }
    return false;
	}
  
  /**
   * This method needs docs.
   */
   
  public Vector getOctetsVector() {
    return octetsVector;
  }
  
  /**
   * Get the size in octets of the specified message.
   *
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #getOctets
   */
	
	public boolean list(int m)
      throws IOException {
    sendCommand("LIST " + m);
    if (getResponse()) {
      int status[] = new int[2];
      status = parseStatus(response.substring(4));
      octets = status[1];
      return true;
    }
    return false;
	}
  
  /**
   * This method needs docs.
   */
   
  public int getOctets() {
    return octets;
  }
  
  /**
   * Read a previously specified message by the retrieve method and
   * write its contents to the specified output stream. 
   * Any byte-stuffed characters will be removed.
   *
   * @param os the output stream to write to
   * @exception IOException if an input or output exception occurred
   * @see #retrieve
   */
	
	public void readTo(OutputStream os)
      throws IOException {
    String s;
    PrintWriter pw = new PrintWriter(os, true);
    
    while ((s = readLine()) != null) {
      pw.println(s);
    }
	}
  
  /**
   * Start the retrieval of the specified message number. This method
   * sends the RETR command to the server. The client needs to repeatedly
   * call <CODE>readLine</CODE> to get the message headers and text.
   *
   * @param message the message number to retrieve
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #readLine
   * @see #readTo
   * @see #getInputStream
   */
   
   public boolean retrieve(int message)
       throws IOException {
     sendCommand("RETR " + message);
     return (getResponse());
   }
   
   /**
    * Get a Pop3InputStream that can be used to retrieve a message
    * from the server. This should only be used after you've called
    * <code>retrieve(int)</code>. Do not use this in conjunction
    * with the <code>readLine</code> method.
    *
    * @return a Pop3InputStream that you can read the message with
    * @exception IOException if an input or output exception occurred
    * @see #retrieve(int)
    * @see #readLine    
    */
   
   public Pop3InputStream getInputStream()
       throws IOException
   {
     return new Pop3InputStream(new BufferedInputStream(socket.getInputStream()));
   }
  
  /**
   * Get a line of text from the server. This method is used after using
   * the <CODE>retrieve</CODE> or <CODE>top</CODE> method. Any
   * byte-stuffed octets are removed.
   *
   * @return a String containing one line of the message being retrieved,
   * or null if it's the end of the message
   * @exception IOException if an input or output exception occurred
   * @see #retrieve
   * @see #top
   */
  
  public String readLine()
      throws IOException {
    String s = in.readLine();
    if (s.length() > 0) {
      if (s.charAt(0) == '.') {
        if (s.length() > 1) {
          s = s.substring(1);
        } else {
          s = null;
        }
      }
    }
    return s;
  }

	/** 
   * Mark the specified message for deletion. This can be undone
   * with the <CODE>reset</CODE> method.
   *
   * @param message the message number to mark for deletion
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #reset
   */
   
   public boolean delete(int message) 
      throws IOException {
    sendCommand("DELE " + message);
    return getResponse();
	}
  
  /**
   * Send the NOOP command to the server. This is used mainly to
   * keep the server's inactivity timer from closing the connection.
   *
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred   
   */
	
	public boolean noOp()
      throws IOException {
    sendCommand("NOOP");
    return getResponse();
	}
  
  /**
   * Restore any messages marked for deletion to their original state.
   *
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #delete
   */
	
	public boolean reset()
      throws IOException {
    sendCommand("RSET");
    return getResponse();
	}
  
  /**
   * Send the QUIT command to the server. Any messages marked
   * for deletion will be expunged at this time. The server should
   * then terminate the connection.
   *
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #delete
   */   
		
	public boolean quit()
      throws IOException {
		sendCommand("QUIT");
    if (getResponse()) {
      state = UPDATE;
      return true;
    }
    return false;
	}
  
  /**
   * Retrieve the headers and specified amount of lines from the
   * specified message. The headers and text will be output to the
   * specified output stream. All byte-stuffed characters are removed
   * from the output. Not all servers implement this command.
   *
   * @param message the message number to retrieve
   * @param lines the number of lines of text the should be retrieved
   * from the body of the message
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   */
	
	public boolean top(int message, int lines, OutputStream os)
      throws IOException {
    String s;
    PrintWriter pw = new PrintWriter(os, true);
    
    sendCommand("TOP " + message + " " + lines);
    if (getResponse()) {
      while ((s = readLine()) != null) {
        pw.println(s);
      }
      return true;
    }
    return false;
  }
  
  /**
   * Retrieve the headers and the specified amount of lines from the body
   * of a message. The <CODE>readLine</CODE> method needs to be called
   * to get each line of the message after this method is called. Not
   * all servers implement this command.
   *
   * @param message the message number to retrieve
   * @param lines the number of lines from the body to retrieve
   * @return a true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #readLine
   */
  
  public boolean top(int m, int n)
      throws IOException {
    sendCommand("TOP " + m + " " + n);
    return (getResponse());
  }
  
  /**
   * Retrieve a vector of strings that can be used to uniquely
   * identify each of the available messages. Messages that are
   * marked for deletion will be represented by an empty string.
   * Note: not all servers implement this command.
   *
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #getUidVector
   */

	public boolean uidl() 
      throws IOException {
    String s;
    StringTokenizer st;
    uidVector = new Vector();
        
    sendCommand("UIDL");
    if (getResponse()) {
      int i = 1;
      do {
        s = getLine();        
        if (!s.equals(".")) {
          st = new StringTokenizer(s);
          int m = Integer.parseInt(st.nextToken());
          if (m != i) {
            for (int j = i; j < m; ++j) {
              uidVector.addElement("");
            }          
          }
          uidVector.addElement(st.nextToken());
          i = m + 1;
        }
      } while (!response.equals("."));
      return true;
    }
    return false;
  }
  
  /**
   * This method needs docs.
   */
   
  public Vector getUidVector() {
    return uidVector;
  }
  
  /**
   * Retrieve a String that can be used to uniquely identify the 
   * specified message. Not all servers implement this command.
   *
   * @param message the message number to retrieve the uid for
   * @return true if the command succeeded, false if not
   * @exception IOException if an input or output exception occurred
   * @see #getUid
   */
  
	public boolean uidl(int m)
      throws IOException {
    sendCommand("UIDL " + m);
    if (getResponse()) {
      uid = response.substring(4);
      uid = uid.substring(uid.indexOf(' '));
      return true;
    }
    return false;
	}
  
  /**
   * This method needs docs.
   */
   
  public String getUid() {
    return uid;
  }   
  
  private static String digestToString(byte[] digest) {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < 16; ++i) {
      if ((digest[i] & 0xFF) < 0x10) {
        sb.append("0");
      }
      sb.append(Integer.toHexString(digest[i] & 0xFF));
    }
    return sb.toString();
  }
  
  /**
   * Closes the connection.
   *
   * @exception IOException if an input or output exception occurred
   */
	
	public void close() 
      throws IOException	{
		out.close();
		in.close();
		socket.close();    
    connected = false;
	}
  
  /**
   * Set whether debug info should be printed to the debug output
   * stream.
   *
   * @param on true if you want debugging on, false if not
   * @see #setDebugOut
   */
  
  public void setDebug(boolean on) {
    debug = on;
  }
  
  /**
   * Set the output stream that debug info should be printed to.
   * The debug bug flag must be set to <CODE>true</CODE> if this
   * setting is to be used.
   *
   * @param os the output stream to print debug info to.
   * @see #setDebug
   */
  
  public void setDebugOut(OutputStream os) {
    debugOut = new PrintWriter(os, true);
  }
}